[Server]
logfile_path = /var/logs/apologies.log
server_host = example.com
server_port = 10
close_timeout_sec = 20
websocket_limit = 30
total_game_limit = 40
in_progress_game_limit = 50
registered_player_limit = 60
websocket_idle_thresh_min = 70
websocket_inactive_thresh_min = 80
player_idle_thresh_min = 90
player_inactive_thresh_min = 100
game_idle_thresh_min = 110
game_inactive_thresh_min = 120
game_retention_thresh_min = 130
idle_websocket_check_period_sec = 140
idle_websocket_check_delay_sec = 150
idle_player_check_period_sec = 160
idle_player_check_delay_sec = 170
idle_game_check_period_sec = 180
idle_game_check_delay_sec = 190
obsolete_game_check_period_sec = 200
obsolete_game_check_delay_sec = 210